"""
Hinting the type using docstring of class/function.

It's an irreplaceable thing if you are using Dependency Injection with passive class:
http://www.martinfowler.com/articles/injection.html

Some code extracted (or based on code) from:
https://github.com/davidhalter/jedi/blob/b489019f5bd5750051122b94cc767df47751ecb7/jedi/evaluate/docstrings.py
Thanks to @davidhalter for this utils under MIT License.

Similar solutions:

    - https://www.jetbrains.com/pycharm/help/type-hinting-in-pycharm.html
    - https://www.python.org/dev/peps/pep-0484/#type-comments
    - http://www.pydev.org/manual_adv_type_hints.html
    - https://jedi.readthedocs.org/en/latest/docs/features.html#type-hinting

Discussions:

    - https://groups.google.com/d/topic/rope-dev/JlAzmZ83K1M/discussion
    - https://groups.google.com/d/topic/rope-dev/LCFNN98vckI/discussion

"""
import re

from rope.base.oi.type_hinting import utils
from rope.base.oi.type_hinting.providers import interfaces


class ParamProvider(interfaces.IParamProvider):
    def __init__(self, docstring_parser, resolver):
        """
        :type docstring_parser: rope.base.oi.type_hinting.providers.docstrings.IParamParser
        :type resolver: rope.base.oi.type_hinting.resolvers.interfaces.IResolver
        """
        self._parse_docstring = docstring_parser
        self._resolve = resolver

    def __call__(self, pyfunc, param_name):
        """
        :type pyfunc: rope.base.pyobjectsdef.PyFunction
        :type param_name: str
        :rtype: rope.base.pyobjects.PyDefinedObject | rope.base.pyobjects.PyObject or None
        """
        type_strs = self._parse_docstring(pyfunc.get_doc(), param_name)
        if type_strs:
            return self._resolve(type_strs[0], pyfunc)


class ReturnProvider(interfaces.IReturnProvider):
    def __init__(self, docstring_parser, resolver):
        """
        :type docstring_parser: rope.base.oi.type_hinting.providers.docstrings.IReturnParser
        :type resolver: rope.base.oi.type_hinting.resolvers.interfaces.IResolver
        """
        self._parse_docstring = docstring_parser
        self._resolve = resolver

    def __call__(self, pyfunc):
        """
        :type pyfunc: rope.base.pyobjectsdef.PyFunction
        :rtype: rope.base.pyobjects.PyDefinedObject | rope.base.pyobjects.PyObject or None
        """
        type_strs = self._parse_docstring(pyfunc.get_doc())
        if type_strs:
            return self._resolve(type_strs[0], pyfunc)


class AssignmentProvider(interfaces.IAssignmentProvider):
    def __init__(self, docstring_parser, resolver):
        """
        :type docstring_parser: rope.base.oi.type_hinting.providers.docstrings.IParamParser
        :type resolver: rope.base.oi.type_hinting.resolvers.interfaces.IResolver
        """
        self._parse_docstring = docstring_parser
        self._resolve = resolver

    def __call__(self, pyname):
        """
        :type pyname: rope.base.pynamesdef.AssignedName
        :rtype: rope.base.pyobjects.PyDefinedObject | rope.base.pyobjects.PyObject or None
        """
        try:
            pyclass, attr_name = utils.get_class_with_attr_name(pyname)
        except TypeError:
            return
        else:
            type_strs = self._parse_docstring(pyclass.get_doc(), attr_name)
            if type_strs:
                return self._resolve(type_strs[0], pyclass)


class IParamParser:
    def __call__(self, docstring, param_name):
        """
        :type docstring: str
        :type param_name: str
        """


class IReturnParser:
    def __call__(self, docstring):
        """
        :type docstring: str
        """


class DocstringParamParser(IParamParser):

    DOCSTRING_PARAM_PATTERNS = [
        r"\s*:type\s+%s:\s*([^\n]+)",  # Sphinx
        r"\s*:param\s+(\w+)\s+%s:[^\n]+",  # Sphinx param with type
        r"\s*@type\s+%s:\s*([^\n]+)",  # Epydoc
    ]

    def __init__(self):
        self._strip_rst_role = RSTRoleStrip()

    def __call__(self, docstring, param_name):
        """Search `docstring` for type(-s) of `param_name`.

        >>> DocstringParamParser()(':type param: int', 'param')
        ['int']
        >>> DocstringParamParser()('@type param: int', 'param')
        ['int']
        >>> DocstringParamParser()(':type param: :class:`threading.Thread`', 'param')
        ['threading.Thread']
        >>> bool(DocstringParamParser()('no document', 'param'))
        False
        >>> DocstringParamParser()(':param int param: some description', 'param')
        ['int']
        """
        if not docstring:
            return []
        patterns = [
            re.compile(p % re.escape(param_name)) for p in self.DOCSTRING_PARAM_PATTERNS
        ]
        for pattern in patterns:
            match = pattern.search(docstring)
            if match:
                return [self._strip_rst_role(match.group(1))]

        return []


class DocstringReturnParser(IReturnParser):

    DOCSTRING_RETURN_PATTERNS = [
        re.compile(r"\s*:rtype:\s*([^\n]+)", re.M),  # Sphinx
        re.compile(r"\s*@rtype:\s*([^\n]+)", re.M),  # Epydoc
    ]

    def __init__(self):
        self._strip_rst_role = RSTRoleStrip()

    def __call__(self, docstring):
        if not docstring:
            return []
        for p in self.DOCSTRING_RETURN_PATTERNS:
            match = p.search(docstring)
            if match:
                return [self._strip_rst_role(match.group(1))]
        return []


class RSTRoleStrip:

    RST_ROLE_PATTERN = re.compile(r":[^`]+:`([^`]+)`")

    def __call__(self, type_str):
        """
        Strip off the part looks like a ReST role in `type_str`.

        >>> RSTRoleStrip()(':class:`ClassName`')  # strip off :class:
        'ClassName'
        >>> RSTRoleStrip()(':py:obj:`module.Object`')  # works with domain
        'module.Object'
        >>> RSTRoleStrip()('ClassName')  # do nothing when not ReST role
        'ClassName'

        See also:
        http://sphinx-doc.org/domains.html#cross-referencing-python-objects

        """
        match = self.RST_ROLE_PATTERN.match(type_str)
        if match:
            return match.group(1)
        else:
            return type_str



import unittest
from textwrap import dedent

import rope.base.builtins  # Use fully-qualified names for clarity.
from rope.base import libutils
from ropetest import testutils


class ObjectInferTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_simple_type_inferencing(self):
        code = dedent("""\
            class Sample(object):
                pass
            a_var = Sample()
        """)
        scope = libutils.get_string_scope(self.project, code)
        sample_class = scope["Sample"].get_object()
        a_var = scope["a_var"].get_object()
        self.assertEqual(sample_class, a_var.get_type())

    def test_simple_type_inferencing_classes_defined_in_holding_scope(self):
        code = dedent("""\
            class Sample(object):
                pass
            def a_func():
                a_var = Sample()
        """)
        scope = libutils.get_string_scope(self.project, code)
        sample_class = scope["Sample"].get_object()
        a_var = scope["a_func"].get_object().get_scope()["a_var"].get_object()
        self.assertEqual(sample_class, a_var.get_type())

    def test_simple_type_inferencing_classes_in_class_methods(self):
        code = dedent("""\
            class Sample(object):
                pass
            class Another(object):
                def a_method():
                    a_var = Sample()
        """)
        scope = libutils.get_string_scope(self.project, code)
        sample_class = scope["Sample"].get_object()
        another_class = scope["Another"].get_object()
        a_var = another_class["a_method"].get_object().get_scope()["a_var"].get_object()
        self.assertEqual(sample_class, a_var.get_type())

    def test_simple_type_inferencing_class_attributes(self):
        code = dedent("""\
            class Sample(object):
                pass
            class Another(object):
                def __init__(self):
                    self.a_var = Sample()
        """)
        scope = libutils.get_string_scope(self.project, code)
        sample_class = scope["Sample"].get_object()
        another_class = scope["Another"].get_object()
        a_var = another_class["a_var"].get_object()
        self.assertEqual(sample_class, a_var.get_type())

    def test_simple_type_inferencing_for_in_class_assignments(self):
        code = dedent("""\
            class Sample(object):
                pass
            class Another(object):
                an_attr = Sample()
        """)
        scope = libutils.get_string_scope(self.project, code)
        sample_class = scope["Sample"].get_object()
        another_class = scope["Another"].get_object()
        an_attr = another_class["an_attr"].get_object()
        self.assertEqual(sample_class, an_attr.get_type())

    def test_simple_type_inferencing_for_chained_assignments(self):
        mod = dedent("""\
            class Sample(object):
                pass
            copied_sample = Sample""")
        mod_scope = libutils.get_string_scope(self.project, mod)
        sample_class = mod_scope["Sample"]
        copied_sample = mod_scope["copied_sample"]
        self.assertEqual(sample_class.get_object(), copied_sample.get_object())

    def test_following_chained_assignments_avoiding_circles(self):
        mod = dedent("""\
            class Sample(object):
                pass
            sample_class = Sample
            sample_class = sample_class
        """)
        mod_scope = libutils.get_string_scope(self.project, mod)
        sample_class = mod_scope["Sample"]
        sample_class_var = mod_scope["sample_class"]
        self.assertEqual(sample_class.get_object(), sample_class_var.get_object())

    def test_function_returned_object_static_type_inference1(self):
        src = dedent("""\
            class Sample(object):
                pass
            def a_func():
                return Sample
            a_var = a_func()
        """)
        scope = libutils.get_string_scope(self.project, src)
        sample_class = scope["Sample"]
        a_var = scope["a_var"]
        self.assertEqual(sample_class.get_object(), a_var.get_object())

    def test_function_returned_object_static_type_inference2(self):
        src = dedent("""\
            class Sample(object):
                pass
            def a_func():
                return Sample()
            a_var = a_func()
        """)
        scope = libutils.get_string_scope(self.project, src)
        sample_class = scope["Sample"].get_object()
        a_var = scope["a_var"].get_object()
        self.assertEqual(sample_class, a_var.get_type())

    def test_recursive_function_returned_object_static_type_inference(self):
        src = dedent("""\
            class Sample(object):
                pass
            def a_func():
                if True:
                    return Sample()
                else:
                    return a_func()
            a_var = a_func()
        """)
        scope = libutils.get_string_scope(self.project, src)
        sample_class = scope["Sample"].get_object()
        a_var = scope["a_var"].get_object()
        self.assertEqual(sample_class, a_var.get_type())

    def test_func_returned_obj_using_call_spec_func_static_type_infer(self):
        src = dedent("""\
            class Sample(object):
                def __call__(self):
                    return Sample
            sample = Sample()
            a_var = sample()""")
        scope = libutils.get_string_scope(self.project, src)
        sample_class = scope["Sample"]
        a_var = scope["a_var"]
        self.assertEqual(sample_class.get_object(), a_var.get_object())

    def test_list_type_inferencing(self):
        src = dedent("""\
            class Sample(object):
                pass
            a_var = [Sample()]
        """)
        scope = libutils.get_string_scope(self.project, src)
        sample_class = scope["Sample"].get_object()
        a_var = scope["a_var"].get_object()
        self.assertNotEqual(sample_class, a_var.get_type())

    def test_attributed_object_inference(self):
        src = dedent("""\
            class Sample(object):
                def __init__(self):
                    self.a_var = None
                def set(self):
                    self.a_var = Sample()
        """)
        scope = libutils.get_string_scope(self.project, src)
        sample_class = scope["Sample"].get_object()
        a_var = sample_class["a_var"].get_object()
        self.assertEqual(sample_class, a_var.get_type())

    def test_getting_property_attributes(self):
        src = dedent("""\
            class A(object):
                pass
            def f(*args):
                return A()
            class B(object):
                p = property(f)
            a_var = B().p
        """)
        pymod = libutils.get_string_module(self.project, src)
        a_class = pymod["A"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(a_class, a_var.get_type())

    def test_getting_property_attributes_with_method_getters(self):
        src = dedent("""\
            class A(object):
                pass
            class B(object):
                def p_get(self):
                    return A()
                p = property(p_get)
            a_var = B().p
        """)
        pymod = libutils.get_string_module(self.project, src)
        a_class = pymod["A"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(a_class, a_var.get_type())

    def test_lambda_functions(self):
        code = dedent("""\
            class C(object):
                pass
            l = lambda: C()
            a_var = l()""")
        mod = libutils.get_string_module(self.project, code)
        c_class = mod["C"].get_object()
        a_var = mod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_mixing_subscript_with_tuple_assigns(self):
        code = dedent("""\
            class C(object):
                attr = 0
            d = {}
            d[0], b = (0, C())
        """)
        mod = libutils.get_string_module(self.project, code)
        c_class = mod["C"].get_object()
        a_var = mod["b"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_mixing_ass_attr_with_tuple_assignment(self):
        code = dedent("""\
            class C(object):
                attr = 0
            c = C()
            c.attr, b = (0, C())
        """)
        mod = libutils.get_string_module(self.project, code)
        c_class = mod["C"].get_object()
        a_var = mod["b"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_mixing_slice_with_tuple_assigns(self):
        code = dedent("""\
            class C(object):
                attr = 0
            d = [None] * 3
            d[0:2], b = ((0,), C())
        """)
        mod = libutils.get_string_module(self.project, code)

        c_class = mod["C"].get_object()
        a_var = mod["b"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_nested_tuple_assignments(self):
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            a, (b, c) = (C1(), (C2(), C1()))
        """)
        mod = libutils.get_string_module(self.project, code)

        c1_class = mod["C1"].get_object()
        c2_class = mod["C2"].get_object()
        a_var = mod["a"].get_object()
        b_var = mod["b"].get_object()
        c_var = mod["c"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())
        self.assertEqual(c1_class, c_var.get_type())

    def test_empty_tuples(self):
        code = dedent("""\
            t = ()
            a, b = t
        """)
        mod = libutils.get_string_module(self.project, code)

        a = mod["a"].get_object()  # noqa

    def test_handling_generator_functions(self):
        code = dedent("""\
            class C(object):
                pass
            def f():
                yield C()
            for c in f():
                a_var = c
        """)
        mod = libutils.get_string_module(self.project, code)
        c_class = mod["C"].get_object()
        a_var = mod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_handling_generator_functions_for_strs(self):
        mod = testutils.create_module(self.project, "mod")
        mod.write(dedent("""\
            def f():
                yield ""
            for s in f():
                a_var = s
        """))
        pymod = self.project.get_pymodule(mod)
        a_var = pymod["a_var"].get_object()
        self.assertTrue(isinstance(a_var.get_type(), rope.base.builtins.Str))

    def test_considering_nones_to_be_unknowns(self):
        code = dedent("""\
            class C(object):
                pass
            a_var = None
            a_var = C()
            a_var = None
        """)
        mod = libutils.get_string_module(self.project, code)
        c_class = mod["C"].get_object()
        a_var = mod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_basic_list_comprehensions(self):
        code = dedent("""\
            class C(object):
                pass
            l = [C() for i in range(1)]
            a_var = l[0]
        """)
        mod = libutils.get_string_module(self.project, code)
        c_class = mod["C"].get_object()
        a_var = mod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_basic_generator_expressions(self):
        code = dedent("""\
            class C(object):
                pass
            l = (C() for i in range(1))
            a_var = list(l)[0]
        """)
        mod = libutils.get_string_module(self.project, code)
        c_class = mod["C"].get_object()
        a_var = mod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_list_comprehensions_and_loop_var(self):
        code = dedent("""\
            class C(object):
                pass
            c_objects = [C(), C()]
            l = [c for c in c_objects]
            a_var = l[0]
        """)
        mod = libutils.get_string_module(self.project, code)
        c_class = mod["C"].get_object()
        a_var = mod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_list_comprehensions_and_multiple_loop_var(self):
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            l = [(c1, c2) for c1 in [C1()] for c2 in [C2()]]
            a, b = l[0]
        """)
        mod = libutils.get_string_module(self.project, code)
        c1_class = mod["C1"].get_object()
        c2_class = mod["C2"].get_object()
        a_var = mod["a"].get_object()
        b_var = mod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_list_comprehensions_and_multiple_iters(self):
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            l = [(c1, c2) for c1, c2 in [(C1(), C2())]]
            a, b = l[0]
        """)
        mod = libutils.get_string_module(self.project, code)

        c1_class = mod["C1"].get_object()
        c2_class = mod["C2"].get_object()
        a_var = mod["a"].get_object()
        b_var = mod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_we_know_the_type_of_caught_exceptions(self):
        code = dedent("""\
            class MyError(Exception):
                pass
            try:

                raise MyError()
            except MyError as e:
                pass
        """)
        mod = libutils.get_string_module(self.project, code)
        my_error = mod["MyError"].get_object()
        e_var = mod["e"].get_object()
        self.assertEqual(my_error, e_var.get_type())

    def test_we_know_the_type_of_caught_multiple_excepts(self):
        code = dedent("""\
            class MyError(Exception):
                pass
            try:
                raise MyError()
            except (MyError, Exception) as e:
                pass
        """)
        mod = libutils.get_string_module(self.project, code)
        my_error = mod["MyError"].get_object()
        e_var = mod["e"].get_object()
        self.assertEqual(my_error, e_var.get_type())

    def test_using_property_as_decorators(self):
        code = dedent("""\
            class A(object):
                pass
            class B(object):
                @property
                def f(self):
                    return A()
            b = B()
            var = b.f
        """)
        mod = libutils.get_string_module(self.project, code)
        var = mod["var"].get_object()
        a = mod["A"].get_object()
        self.assertEqual(a, var.get_type())

    def test_using_property_as_decorators_and_passing_parameter(self):
        code = dedent("""\
            class B(object):
                @property
                def f(self):
                    return self
            b = B()
            var = b.f
        """)
        mod = libutils.get_string_module(self.project, code)
        var = mod["var"].get_object()
        a = mod["B"].get_object()
        self.assertEqual(a, var.get_type())



import rope.refactor.importutils
from rope.base.change import ChangeContents, ChangeSet, CreateFolder, MoveResource


class ModuleToPackage:
    def __init__(self, project, resource):
        self.project = project
        self.resource = resource

    def get_changes(self):
        changes = ChangeSet("Transform <%s> module to package" % self.resource.path)
        new_content = self._transform_relatives_to_absolute(self.resource)
        if new_content is not None:
            changes.add_change(ChangeContents(self.resource, new_content))
        parent = self.resource.parent
        name = self.resource.name[:-3]
        changes.add_change(CreateFolder(parent, name))
        parent_path = parent.path + "/"
        if not parent.path:
            parent_path = ""
        new_path = parent_path + "%s/__init__.py" % name
        if self.resource.project == self.project:
            changes.add_change(MoveResource(self.resource, new_path))
        return changes

    def _transform_relatives_to_absolute(self, resource):
        pymodule = self.project.get_pymodule(resource)
        import_tools = rope.refactor.importutils.ImportTools(self.project)
        return import_tools.relatives_to_absolutes(pymodule)



import inspect
import os
import sys

from rope.base import utils


def _stdlib_path():

    return os.path.dirname(inspect.getsourcefile(inspect))


@utils.cached(1)
def standard_modules():
    return python_modules() | dynload_modules()


@utils.cached(1)
def python_modules():
    result = set()
    lib_path = _stdlib_path()
    if os.path.exists(lib_path):
        for name in os.listdir(lib_path):
            path = os.path.join(lib_path, name)
            if os.path.isdir(path):
                if "-" not in name:
                    result.add(name)
            else:
                if name.endswith(".py"):
                    result.add(name[:-3])
    return result


def normalize_so_name(name):
    """
    Handle different types of python installations
    """
    if "cpython" in name:
        return os.path.splitext(os.path.splitext(name)[0])[0]
    # XXX: Special handling for Fedora python2 distribution
    # See: https://github.com/python-rope/rope/issues/211
    if name == "timemodule.so":
        return "time"
    return os.path.splitext(name)[0]


@utils.cached(1)
def dynload_modules():
    result = set(sys.builtin_module_names)
    dynload_path = os.path.join(_stdlib_path(), "lib-dynload")
    if os.path.exists(dynload_path):
        for name in os.listdir(dynload_path):
            path = os.path.join(dynload_path, name)
            if os.path.isfile(path):
                if name.endswith(".dll"):
                    result.add(normalize_so_name(name))
                if name.endswith(".so"):
                    result.add(normalize_so_name(name))
    return result



from rope.base import pyobjects, worder
from rope.base.builtins import Lambda


class DefinitionInfo:
    def __init__(
        self, function_name, is_method, args_with_defaults, args_arg, keywords_arg
    ):
        self.function_name = function_name
        self.is_method = is_method
        self.args_with_defaults = args_with_defaults
        self.args_arg = args_arg
        self.keywords_arg = keywords_arg

    def to_string(self):
        return f"{self.function_name}({self.arguments_to_string()})"

    def arguments_to_string(self, from_index=0):
        params = []
        for arg, default in self.args_with_defaults:
            if default is not None:
                params.append(f"{arg}={default}")
            else:
                params.append(arg)
        if self.args_arg is not None:
            params.append("*" + self.args_arg)
        if self.keywords_arg:
            params.append("**" + self.keywords_arg)
        return ", ".join(params[from_index:])

    @staticmethod
    def _read(pyfunction, code):
        kind = pyfunction.get_kind()
        is_method = kind == "method"
        is_lambda = kind == "lambda"
        info = _FunctionParser(code, is_method, is_lambda)
        args, keywords = info.get_parameters()
        args_arg = None
        keywords_arg = None
        if args and args[-1].startswith("**"):
            keywords_arg = args[-1][2:]
            del args[-1]
        if args and args[-1].startswith("*"):
            args_arg = args[-1][1:]
            del args[-1]
        args_with_defaults = [(name, None) for name in args]
        args_with_defaults.extend(keywords)
        return DefinitionInfo(
            info.get_function_name(),
            is_method,
            args_with_defaults,
            args_arg,
            keywords_arg,
        )

    @staticmethod
    def read(pyfunction):
        pymodule = pyfunction.get_module()
        word_finder = worder.Worder(pymodule.source_code)
        lineno = pyfunction.get_ast().lineno
        start = pymodule.lines.get_line_start(lineno)
        if isinstance(pyfunction, Lambda):
            call = word_finder.get_lambda_and_args(start)
        else:
            call = word_finder.get_function_and_args_in_header(start)
        return DefinitionInfo._read(pyfunction, call)


class CallInfo:
    def __init__(
        self,
        function_name,
        args,
        keywords,
        args_arg,
        keywords_arg,
        implicit_arg,
        constructor,
    ):
        self.function_name = function_name
        self.args = args
        self.keywords = keywords
        self.args_arg = args_arg
        self.keywords_arg = keywords_arg
        self.implicit_arg = implicit_arg
        self.constructor = constructor

    def to_string(self):
        function = self.function_name
        if self.implicit_arg:
            function = self.args[0] + "." + self.function_name
        params = []
        start = 0
        if self.implicit_arg or self.constructor:
            start = 1
        if self.args[start:]:
            params.extend(self.args[start:])
        if self.keywords:
            params.extend([f"{name}={value}" for name, value in self.keywords])
        if self.args_arg is not None:
            params.append("*" + self.args_arg)
        if self.keywords_arg:
            params.append("**" + self.keywords_arg)
        return "{}({})".format(function, ", ".join(params))

    @staticmethod
    def read(primary, pyname, definition_info, code):
        is_method_call = CallInfo._is_method_call(primary, pyname)
        is_constructor = CallInfo._is_class(pyname)
        is_classmethod = CallInfo._is_classmethod(pyname)
        info = _FunctionParser(code, is_method_call or is_classmethod)
        args, keywords = info.get_parameters()
        args_arg = None
        keywords_arg = None
        if args and args[-1].startswith("**"):
            keywords_arg = args[-1][2:]
            del args[-1]
        if args and args[-1].startswith("*"):
            args_arg = args[-1][1:]
            del args[-1]
        if is_constructor:
            args.insert(0, definition_info.args_with_defaults[0][0])
        return CallInfo(
            info.get_function_name(),
            args,
            keywords,
            args_arg,
            keywords_arg,
            is_method_call or is_classmethod,
            is_constructor,
        )

    @staticmethod
    def _is_method_call(primary, pyname):
        return (
            primary is not None
            and isinstance(primary.get_object().get_type(), pyobjects.PyClass)
            and CallInfo._is_method(pyname)
        )

    @staticmethod
    def _is_class(pyname):
        return pyname is not None and isinstance(pyname.get_object(), pyobjects.PyClass)

    @staticmethod
    def _is_method(pyname):
        if pyname is not None and isinstance(pyname.get_object(), pyobjects.PyFunction):
            return pyname.get_object().get_kind() == "method"
        return False

    @staticmethod
    def _is_classmethod(pyname):
        if pyname is not None and isinstance(pyname.get_object(), pyobjects.PyFunction):
            return pyname.get_object().get_kind() == "classmethod"
        return False


class ArgumentMapping:
    def __init__(self, definition_info, call_info):
        self.call_info = call_info
        self.param_dict = {}
        self.keyword_args = []
        self.args_arg = []
        for index, value in enumerate(call_info.args):
            if index < len(definition_info.args_with_defaults):
                name = definition_info.args_with_defaults[index][0]
                self.param_dict[name] = value
            else:
                self.args_arg.append(value)
        for name, value in call_info.keywords:
            index = -1
            for pair in definition_info.args_with_defaults:
                if pair[0] == name:
                    self.param_dict[name] = value
                    break
            else:
                self.keyword_args.append((name, value))

    def to_call_info(self, definition_info):
        args = []
        keywords = []
        for index in range(len(definition_info.args_with_defaults)):
            name = definition_info.args_with_defaults[index][0]
            if name in self.param_dict:
                args.append(self.param_dict[name])
            else:
                for i in range(index, len(definition_info.args_with_defaults)):
                    name = definition_info.args_with_defaults[i][0]
                    if name in self.param_dict:
                        keywords.append((name, self.param_dict[name]))
                break
        args.extend(self.args_arg)
        keywords.extend(self.keyword_args)
        return CallInfo(
            self.call_info.function_name,
            args,
            keywords,
            self.call_info.args_arg,
            self.call_info.keywords_arg,
            self.call_info.implicit_arg,
            self.call_info.constructor,
        )


class _FunctionParser:
    def __init__(self, call, implicit_arg, is_lambda=False):
        self.call = call
        self.implicit_arg = implicit_arg
        self.word_finder = worder.Worder(self.call)
        if is_lambda:
            self.last_parens = self.call.rindex(":")
        else:
            self.last_parens = self.call.rindex(")")
        self.first_parens = self.word_finder._find_parens_start(self.last_parens)

    def get_parameters(self):
        args, keywords = self.word_finder.get_parameters(
            self.first_parens, self.last_parens
        )
        if self.is_called_as_a_method():
            instance = self.call[: self.call.rindex(".", 0, self.first_parens)]
            args.insert(0, instance.strip())
        return args, keywords

    def get_instance(self):
        if self.is_called_as_a_method():
            return self.word_finder.get_primary_at(
                self.call.rindex(".", 0, self.first_parens) - 1
            )

    def get_function_name(self):
        if self.is_called_as_a_method():
            return self.word_finder.get_word_at(self.first_parens - 1)
        else:
            return self.word_finder.get_primary_at(self.first_parens - 1)

    def is_called_as_a_method(self):
        return self.implicit_arg and "." in self.call[: self.first_parens]



# Special cases, easier to express in pytest
from contextlib import closing
from textwrap import dedent

import pytest

from rope.base.project import Project
from rope.base.resources import File, Folder
from rope.contrib.autoimport.sqlite import AutoImport


@pytest.fixture
def autoimport(project: Project):
    with closing(AutoImport(project)) as ai:
        yield ai


def test_init_py(
    autoimport: AutoImport,
    project: Project,
    pkg1: Folder,
    mod1: File,
):
    mod1_init = pkg1.get_child("__init__.py")
    mod1_init.write(dedent("""\
        def foo():
            pass
    """))
    mod1.write(dedent("""\
        foo
    """))
    autoimport.generate_cache([mod1_init])
    results = autoimport.search("foo", True)
    assert [("from pkg1 import foo", "foo")] == results



"""This module can be used for finding similar code"""
import re

import rope.base.builtins  # Use full qualification for clarity.
import rope.refactor.wildcards  # Use full qualification for clarity.
from rope.base import ast, codeanalyze, exceptions, libutils
from rope.refactor import patchedast
from rope.refactor.patchedast import MismatchedTokenError


class BadNameInCheckError(exceptions.RefactoringError):
    pass


class SimilarFinder:
    """`SimilarFinder` can be used to find similar pieces of code

    See the notes in the `rope.refactor.restructure` module for more
    info.

    """

    def __init__(self, pymodule, wildcards=None):
        """Construct a SimilarFinder"""
        self.source = pymodule.source_code
        try:
            self.raw_finder = RawSimilarFinder(
                pymodule.source_code, pymodule.get_ast(), self._does_match
            )
        except MismatchedTokenError:
            print("in file %s" % pymodule.resource.path)
            raise
        self.pymodule = pymodule
        if wildcards is None:
            self.wildcards = {}
            for wildcard in [
                rope.refactor.wildcards.DefaultWildcard(pymodule.pycore.project)
            ]:
                self.wildcards[wildcard.get_name()] = wildcard
        else:
            self.wildcards = wildcards

    def get_matches(self, code, args=None, start=0, end=None):
        if args is None:
            args = {}
        self.args = args
        if end is None:
            end = len(self.source)
        skip_region = None
        if "skip" in args.get("", {}):
            resource, region = args[""]["skip"]
            if resource == self.pymodule.get_resource():
                skip_region = region
        return self.raw_finder.get_matches(code, start=start, end=end, skip=skip_region)

    def get_match_regions(self, *args, **kwds):
        for match in self.get_matches(*args, **kwds):
            yield match.get_region()

    def _does_match(self, node, name):
        arg = self.args.get(name, "")
        kind = "default"
        if isinstance(arg, (tuple, list)):
            kind = arg[0]
            arg = arg[1]
        suspect = rope.refactor.wildcards.Suspect(self.pymodule, node, name)
        return self.wildcards[kind].matches(suspect, arg)


class RawSimilarFinder:
    """A class for finding similar expressions and statements"""

    def __init__(self, source, node=None, does_match=None):
        if node is None:
            try:
                node = ast.parse(source)
            except SyntaxError:
                # needed to parse expression containing := operator
                node = ast.parse("(" + source + ")")
        if does_match is None:
            self.does_match = self._simple_does_match
        else:
            self.does_match = does_match
        self._init_using_ast(node, source)

    def _simple_does_match(self, node, name):
        return isinstance(node, (ast.expr, ast.Name))

    def _init_using_ast(self, node, source):
        self.source = source
        self._matched_asts = {}
        if not hasattr(node, "region"):
            patchedast.patch_ast(node, source)
        self.ast = node

    def get_matches(self, code, start=0, end=None, skip=None):
        """Search for `code` in source and return a list of `Match`-es

        `code` can contain wildcards.  ``${name}`` matches normal
        names and ``${?name} can match any expression.  You can use
        `Match.get_ast()` for getting the node that has matched a
        given pattern.

        """
        if end is None:
            end = len(self.source)
        for match in self._get_matched_asts(code):
            match_start, match_end = match.get_region()
            if start <= match_start and match_end <= end:
                if skip is not None and (skip[0] < match_end and skip[1] > match_start):
                    continue
                yield match

    def _get_matched_asts(self, code):
        if code not in self._matched_asts:
            wanted = self._create_pattern(code)
            matches = _ASTMatcher(self.ast, wanted, self.does_match).find_matches()
            self._matched_asts[code] = matches
        return self._matched_asts[code]

    def _create_pattern(self, expression):
        expression = self._replace_wildcards(expression)
        node = ast.parse(expression)
        # Getting Module.Stmt.nodes
        nodes = node.body
        if len(nodes) == 1 and isinstance(nodes[0], ast.Expr):
            # Getting Discard.expr
            wanted = nodes[0].value
        else:
            wanted = nodes
        return wanted

    def _replace_wildcards(self, expression):
        ropevar = _RopeVariable()
        template = CodeTemplate(expression)
        mapping = {name: ropevar.get_var(name) for name in template.get_names()}
        return template.substitute(mapping)


class _ASTMatcher:
    def __init__(self, body, pattern, does_match):
        """Searches the given pattern in the body AST.

        body is an AST node and pattern can be either an AST node or
        a list of ASTs nodes
        """
        self.body = body
        self.pattern = pattern
        self.matches = None
        self.ropevar = _RopeVariable()
        self.matches_callback = does_match

    def find_matches(self):
        if self.matches is None:
            self.matches = []
            # _check_nodes always returns None, so
            # call_for_nodes traverses self.body's entire tree.
            ast.call_for_nodes(self.body, self._check_node)
        return self.matches

    def _check_node(self, node):
        if isinstance(self.pattern, list):
            self._check_statements(node)
        else:
            self._check_expression(node)

    def _check_expression(self, node):
        mapping = {}
        if self._match_nodes(self.pattern, node, mapping):
            self.matches.append(ExpressionMatch(node, mapping))

    def _check_statements(self, node):
        for field, child in ast.iter_fields(node):
            if isinstance(child, (list, tuple)):
                self.__check_stmt_list(child)

    def __check_stmt_list(self, nodes):
        for index in range(len(nodes)):
            if len(nodes) - index >= len(self.pattern):
                current_stmts = nodes[index : index + len(self.pattern)]
                mapping = {}
                if self._match_stmts(current_stmts, mapping):
                    self.matches.append(StatementMatch(current_stmts, mapping))

    def _match_nodes(self, expected, node, mapping):
        if isinstance(expected, ast.Name):
            if self.ropevar.is_var(expected.id):
                return self._match_wildcard(expected, node, mapping)
        if not isinstance(expected, ast.AST):
            return expected == node
        if expected.__class__ != node.__class__:
            return False

        children1 = self._get_children(expected)
        children2 = self._get_children(node)
        if len(children1) != len(children2):
            return False
        for child1, child2 in zip(children1, children2):
            if isinstance(child1, ast.AST):
                if not self._match_nodes(child1, child2, mapping):
                    return False
            elif isinstance(child1, (list, tuple)):
                if not isinstance(child2, (list, tuple)) or len(child1) != len(child2):
                    return False
                for c1, c2 in zip(child1, child2):
                    if not self._match_nodes(c1, c2, mapping):
                        return False
            else:
                if type(child1) is not type(child2) or child1 != child2:
                    return False
        return True

    def _get_children(self, node):
        """Return not `ast.expr_context` children of `node`"""
        return [
            child
            for field, child in ast.iter_fields(node)
            if not isinstance(child, ast.expr_context)
        ]

    def _match_stmts(self, current_stmts, mapping):
        if len(current_stmts) != len(self.pattern):
            return False
        for stmt, expected in zip(current_stmts, self.pattern):
            if not self._match_nodes(expected, stmt, mapping):
                return False
        return True

    def _match_wildcard(self, node1, node2, mapping):
        name = self.ropevar.get_base(node1.id)
        if name not in mapping:
            if self.matches_callback(node2, name):
                mapping[name] = node2
                return True
            return False
        else:
            return self._match_nodes(mapping[name], node2, {})


class Match:
    def __init__(self, mapping):
        self.mapping = mapping

    def get_region(self):
        """Returns match region"""

    def get_ast(self, name):
        """Return the ast node that has matched rope variables"""
        return self.mapping.get(name, None)


class ExpressionMatch(Match):
    def __init__(self, ast, mapping):
        super().__init__(mapping)
        self.ast = ast

    def get_region(self):
        return self.ast.region


class StatementMatch(Match):
    def __init__(self, ast_list, mapping):
        super().__init__(mapping)
        self.ast_list = ast_list

    def get_region(self):
        return self.ast_list[0].region[0], self.ast_list[-1].region[1]


class CodeTemplate:
    def __init__(self, template):
        self.template = template
        self._find_names()

    def _find_names(self):
        self.names = {}
        for match in CodeTemplate._get_pattern().finditer(self.template):
            if "name" in match.groupdict() and match.group("name") is not None:
                start, end = match.span("name")
                name = self.template[start + 2 : end - 1]
                if name not in self.names:
                    self.names[name] = []
                self.names[name].append((start, end))

    def get_names(self):
        return self.names.keys()

    def substitute(self, mapping):
        collector = codeanalyze.ChangeCollector(self.template)
        for name, occurrences in self.names.items():
            for region in occurrences:
                collector.add_change(region[0], region[1], mapping[name])
        result = collector.get_changed()
        if result is None:
            return self.template
        return result

    _match_pattern = None

    @classmethod
    def _get_pattern(cls):
        if cls._match_pattern is None:
            pattern = (
                codeanalyze.get_comment_pattern()
                + "|"
                + codeanalyze.get_string_pattern()
                + "|"
                + r"(?P<name>\$\{[^\s\$\}]*\})"
            )
            cls._match_pattern = re.compile(pattern)
        return cls._match_pattern


class _RopeVariable:
    """Transform and identify rope inserted wildcards"""

    _normal_prefix = "__rope__variable_normal_"
    _any_prefix = "__rope__variable_any_"

    def get_var(self, name):
        if name.startswith("?"):
            return self._get_any(name)
        else:
            return self._get_normal(name)

    def is_var(self, name):
        return self._is_normal(name) or self._is_var(name)

    def get_base(self, name):
        if self._is_normal(name):
            return name[len(self._normal_prefix) :]
        if self._is_var(name):
            return "?" + name[len(self._any_prefix) :]

    def _get_normal(self, name):
        return self._normal_prefix + name

    def _get_any(self, name):
        return self._any_prefix + name[1:]

    def _is_normal(self, name):
        return name.startswith(self._normal_prefix)

    def _is_var(self, name):
        return name.startswith(self._any_prefix)


def make_pattern(code, variables):
    variables = set(variables)
    collector = codeanalyze.ChangeCollector(code)

    def does_match(node, name):
        return isinstance(node, ast.Name) and node.id == name

    finder = RawSimilarFinder(code, does_match=does_match)
    for variable in variables:
        for match in finder.get_matches("${%s}" % variable):
            start, end = match.get_region()
            collector.add_change(start, end, "${%s}" % variable)
    result = collector.get_changed()
    return result if result is not None else code


def _pydefined_to_str(pydefined):
    address = []
    if isinstance(
        pydefined, (rope.base.builtins.BuiltinClass, rope.base.builtins.BuiltinFunction)
    ):
        return f"__builtins__.{pydefined.get_name()}"
    while pydefined.parent is not None:
        address.insert(0, pydefined.get_name())
        pydefined = pydefined.parent
    module_name = libutils.modname(pydefined.resource)
    return ".".join(module_name.split(".") + address)



"""This module tries to support builtin types and functions."""
import inspect
import io

import rope.base.evaluate
from rope.base import arguments, ast, pynames, pyobjects, utils


class BuiltinModule(pyobjects.AbstractModule):
    def __init__(self, name, pycore=None, initial={}):
        super().__init__()
        self.name = name
        self.pycore = pycore
        self.initial = initial

    parent = None

    def get_attributes(self):
        return self.attributes

    def get_doc(self):
        if self.module:
            return self.module.__doc__

    def get_name(self):
        return self.name.split(".")[-1]

    @property
    @utils.saveit
    def attributes(self):
        result = _object_attributes(self.module, self)
        result.update(self.initial)
        if self.pycore is not None:
            submodules = self.pycore._builtin_submodules(self.name)
            for name, module in submodules.items():
                result[name] = rope.base.builtins.BuiltinName(module)
        return result

    @property
    @utils.saveit
    def module(self):
        try:
            result = __import__(self.name)
            for token in self.name.split(".")[1:]:
                result = getattr(result, token, None)
            return result
        except ImportError:
            return


class _BuiltinElement:
    def __init__(self, builtin, parent=None):
        self.builtin = builtin
        self._parent = parent

    def get_doc(self):
        if self.builtin:
            return getattr(self.builtin, "__doc__", None)

    def get_name(self):
        if self.builtin:
            return getattr(self.builtin, "__name__", None)

    @property
    def parent(self):
        if self._parent is None:
            return builtins
        return self._parent


class BuiltinClass(_BuiltinElement, pyobjects.AbstractClass):
    def __init__(self, builtin, attributes, parent=None):
        _BuiltinElement.__init__(self, builtin, parent)
        pyobjects.AbstractClass.__init__(self)
        self.initial = attributes

    @utils.saveit
    def get_attributes(self):
        result = _object_attributes(self.builtin, self)
        result.update(self.initial)
        return result

    def get_module(self):
        return builtins


class BuiltinFunction(_BuiltinElement, pyobjects.AbstractFunction):
    def __init__(
        self, returned=None, function=None, builtin=None, argnames=[], parent=None
    ):
        _BuiltinElement.__init__(self, builtin, parent)
        pyobjects.AbstractFunction.__init__(self)
        self.argnames = argnames
        self.returned = returned
        self.function = function

    def get_returned_object(self, args):
        if self.function is not None:
            return self.function(_CallContext(self.argnames, args))
        else:
            return self.returned

    def get_param_names(self, special_args=True):
        return self.argnames


class BuiltinUnknown(_BuiltinElement, pyobjects.PyObject):
    def __init__(self, builtin):
        super().__init__(pyobjects.get_unknown())
        self.builtin = builtin
        self.type = pyobjects.get_unknown()

    def get_name(self):
        return getattr(type(self.builtin), "__name__", None)

    @utils.saveit
    def get_attributes(self):
        return _object_attributes(self.builtin, self)


def _object_attributes(obj, parent):
    attributes = {}
    for name in dir(obj):
        if name == "None":
            continue
        try:
            child = getattr(obj, name)
        except AttributeError:
            # descriptors are allowed to raise AttributeError
            # even if they are in dir()
            continue
        pyobject = None
        if inspect.isclass(child):
            pyobject = BuiltinClass(child, {}, parent=parent)
        elif inspect.isroutine(child):
            pyobject = BuiltinFunction(builtin=child, parent=parent)
        else:
            pyobject = BuiltinUnknown(builtin=child)
        attributes[name] = BuiltinName(pyobject)
    return attributes


def _create_builtin_type_getter(cls):
    def _get_builtin(*args):
        if not hasattr(cls, "_generated"):
            cls._generated = {}
        if args not in cls._generated:
            cls._generated[args] = cls(*args)
        return cls._generated[args]

    return _get_builtin


def _create_builtin_getter(cls):
    type_getter = _create_builtin_type_getter(cls)

    def _get_builtin(*args):
        return pyobjects.PyObject(type_getter(*args))

    return _get_builtin


class _CallContext:
    def __init__(self, argnames, args):
        self.argnames = argnames
        self.args = args

    def _get_scope_and_pyname(self, pyname):
        if pyname is not None and isinstance(pyname, pynames.AssignedName):
            pymodule, lineno = pyname.get_definition_location()
            if pymodule is None:
                return None, None
            if lineno is None:
                lineno = 1
            scope = pymodule.get_scope().get_inner_scope_for_line(lineno)
            name = None
            while name is None and scope is not None:
                for current in scope.get_names():
                    if scope[current] is pyname:
                        name = current
                        break
                else:
                    scope = scope.parent
            return scope, name
        return None, None

    def get_argument(self, name):
        if self.args:
            args = self.args.get_arguments(self.argnames)
            return args[self.argnames.index(name)]

    def get_pyname(self, name):
        if self.args:
            args = self.args.get_pynames(self.argnames)
            if name in self.argnames:
                return args[self.argnames.index(name)]

    def get_arguments(self, argnames):
        if self.args:
            return self.args.get_arguments(argnames)

    def get_pynames(self, argnames):
        if self.args:
            return self.args.get_pynames(argnames)

    def get_per_name(self):
        if self.args is None:
            return None
        pyname = self.args.get_instance_pyname()
        scope, name = self._get_scope_and_pyname(pyname)
        if name is not None:
            pymodule = pyname.get_definition_location()[0]
            return pymodule.pycore.object_info.get_per_name(scope, name)
        return None

    def save_per_name(self, value):
        if self.args is None:
            return None
        pyname = self.args.get_instance_pyname()
        scope, name = self._get_scope_and_pyname(pyname)
        if name is not None:
            pymodule = pyname.get_definition_location()[0]
            pymodule.pycore.object_info.save_per_name(scope, name, value)


class _AttributeCollector:
    def __init__(self, type):
        self.attributes = {}
        self.type = type

    def __call__(
        self,
        name,
        returned=None,
        function=None,
        argnames=["self"],
        check_existence=True,
        parent=None,
    ):
        try:
            builtin = getattr(self.type, name)
        except AttributeError:
            if check_existence:
                raise
            builtin = None
        self.attributes[name] = BuiltinName(
            BuiltinFunction(
                returned=returned,
                function=function,
                argnames=argnames,
                builtin=builtin,
                parent=parent,
            )
        )

    def __setitem__(self, name, value):
        self.attributes[name] = value


class List(BuiltinClass):
    def __init__(self, holding=None):
        self.holding = holding
        collector = _AttributeCollector(list)

        collector("__iter__", function=self._iterator_get, parent=self)
        collector("__new__", function=self._new_list, parent=self)

        # Adding methods
        collector(
            "append", function=self._list_add, argnames=["self", "value"], parent=self
        )
        collector(
            "__setitem__",
            function=self._list_add,
            argnames=["self", "index", "value"],
            parent=self,
        )
        collector(
            "insert",
            function=self._list_add,
            argnames=["self", "index", "value"],
            parent=self,
        )
        collector(
            "extend",
            function=self._self_set,
            argnames=["self", "iterable"],
            parent=self,
        )

        # Getting methods
        collector("__getitem__", function=self._list_get, parent=self)
        collector("pop", function=self._list_get, parent=self)
        try:
            collector("__getslice__", function=self._list_get)
        except AttributeError:
            pass

        super().__init__(list, collector.attributes)

    def _new_list(self, args):
        return _create_builtin(args, get_list)

    def _list_add(self, context):
        if self.holding is not None:
            return
        holding = context.get_argument("value")
        if holding is not None and holding != pyobjects.get_unknown():
            context.save_per_name(holding)

    def _self_set(self, context):
        if self.holding is not None:
            return
        iterable = context.get_pyname("iterable")
        holding = _infer_sequence_for_pyname(iterable)
        if holding is not None and holding != pyobjects.get_unknown():
            context.save_per_name(holding)

    def _list_get(self, context):
        if self.holding is not None:
            args = context.get_arguments(["self", "key"])
            if (
                len(args) > 1
                and args[1] is not None
                and args[1].get_type() == builtins["slice"].get_object()
            ):
                return get_list(self.holding)
            return self.holding
        return context.get_per_name()

    def _iterator_get(self, context):
        return get_iterator(self._list_get(context))

    def _self_get(self, context):
        return get_list(self._list_get(context))


get_list = _create_builtin_getter(List)
get_list_type = _create_builtin_type_getter(List)


class Dict(BuiltinClass):
    def __init__(self, keys=None, values=None):
        self.keys = keys
        self.values = values
        collector = _AttributeCollector(dict)
        collector("__new__", function=self._new_dict, parent=self)
        collector("__setitem__", function=self._dict_add, parent=self)
        collector("popitem", function=self._item_get, parent=self)
        collector("pop", function=self._value_get, parent=self)
        collector("get", function=self._key_get, parent=self)
        collector("keys", function=self._key_list, parent=self)
        collector("values", function=self._value_list, parent=self)
        collector("items", function=self._item_list, parent=self)
        collector("copy", function=self._self_get, parent=self)
        collector("__getitem__", function=self._value_get, parent=self)
        collector("__iter__", function=self._key_iter, parent=self)
        collector("update", function=self._self_set, parent=self)
        super().__init__(dict, collector.attributes)

    def _new_dict(self, args):
        def do_create(holding=None):
            if holding is None:
                return get_dict()
            type = holding.get_type()
            if isinstance(type, Tuple) and len(type.get_holding_objects()) == 2:
                return get_dict(*type.get_holding_objects())

        return _create_builtin(args, do_create)

    def _dict_add(self, context):
        if self.keys is not None:
            return
        key, value = context.get_arguments(["self", "key", "value"])[1:]
        if key is not None and key != pyobjects.get_unknown():
            context.save_per_name(get_tuple(key, value))

    def _item_get(self, context):
        if self.keys is not None:
            return get_tuple(self.keys, self.values)
        item = context.get_per_name()
        if item is None or not isinstance(item.get_type(), Tuple):
            return get_tuple(self.keys, self.values)
        return item

    def _value_get(self, context):
        item = self._item_get(context).get_type()
        return item.get_holding_objects()[1]

    def _key_get(self, context):
        item = self._item_get(context).get_type()
        return item.get_holding_objects()[0]

    def _value_list(self, context):
        return get_list(self._value_get(context))

    def _key_list(self, context):
        return get_list(self._key_get(context))

    def _item_list(self, context):
        return get_list(self._item_get(context))

    def _value_iter(self, context):
        return get_iterator(self._value_get(context))

    def _key_iter(self, context):
        return get_iterator(self._key_get(context))

    def _item_iter(self, context):
        return get_iterator(self._item_get(context))

    def _self_get(self, context):
        item = self._item_get(context).get_type()
        key, value = item.get_holding_objects()[:2]
        return get_dict(key, value)

    def _self_set(self, context):
        if self.keys is not None:
            return
        new_dict = context.get_pynames(["self", "d"])[1]
        if new_dict and isinstance(new_dict.get_object().get_type(), Dict):
            args = arguments.ObjectArguments([new_dict])
            items = (
                new_dict.get_object()["popitem"].get_object().get_returned_object(args)
            )
            context.save_per_name(items)
        else:
            holding = _infer_sequence_for_pyname(new_dict)
            if holding is not None and isinstance(holding.get_type(), Tuple):
                context.save_per_name(holding)


get_dict = _create_builtin_getter(Dict)
get_dict_type = _create_builtin_type_getter(Dict)


class Tuple(BuiltinClass):
    def __init__(self, *objects):
        self.objects = objects
        first = None
        if objects:
            first = objects[0]
        attributes = {
            "__getitem__": BuiltinName(
                BuiltinFunction(first)
            ),  # TODO: add slice support
            "__getslice__": BuiltinName(BuiltinFunction(pyobjects.PyObject(self))),
            "__new__": BuiltinName(BuiltinFunction(function=self._new_tuple)),
            "__iter__": BuiltinName(BuiltinFunction(get_iterator(first))),
        }
        super().__init__(tuple, attributes)

    def get_holding_objects(self):
        return self.objects

    def _new_tuple(self, args):
        return _create_builtin(args, get_tuple)


get_tuple = _create_builtin_getter(Tuple)
get_tuple_type = _create_builtin_type_getter(Tuple)


class Set(BuiltinClass):
    def __init__(self, holding=None):
        self.holding = holding
        collector = _AttributeCollector(set)
        collector("__new__", function=self._new_set)

        self_methods = [
            "copy",
            "difference",
            "intersection",
            "symmetric_difference",
            "union",
        ]
        for method in self_methods:
            collector(method, function=self._self_get, parent=self)
        collector("add", function=self._set_add, parent=self)
        collector("update", function=self._self_set, parent=self)
        collector("update", function=self._self_set, parent=self)
        collector("symmetric_difference_update", function=self._self_set, parent=self)
        collector("difference_update", function=self._self_set, parent=self)

        collector("pop", function=self._set_get, parent=self)
        collector("__iter__", function=self._iterator_get, parent=self)
        super().__init__(set, collector.attributes)

    def _new_set(self, args):
        return _create_builtin(args, get_set)

    def _set_add(self, context):
        if self.holding is not None:
            return
        holding = context.get_arguments(["self", "value"])[1]
        if holding is not None and holding != pyobjects.get_unknown():
            context.save_per_name(holding)

    def _self_set(self, context):
        if self.holding is not None:
            return
        iterable = context.get_pyname("iterable")
        holding = _infer_sequence_for_pyname(iterable)
        if holding is not None and holding != pyobjects.get_unknown():
            context.save_per_name(holding)

    def _set_get(self, context):
        if self.holding is not None:
            return self.holding
        return context.get_per_name()

    def _iterator_get(self, context):
        return get_iterator(self._set_get(context))

    def _self_get(self, context):
        return get_list(self._set_get(context))


get_set = _create_builtin_getter(Set)
get_set_type = _create_builtin_type_getter(Set)


class Str(BuiltinClass):
    def __init__(self):
        self_object = pyobjects.PyObject(self)
        collector = _AttributeCollector(str)
        collector("__iter__", get_iterator(self_object), check_existence=False)

        self_methods = [
            "__getitem__",
            "capitalize",
            "center",
            "encode",
            "expandtabs",
            "join",
            "ljust",
            "lower",
            "lstrip",
            "replace",
            "rjust",
            "rstrip",
            "strip",
            "swapcase",
            "title",
            "translate",
            "upper",
            "zfill",
        ]
        for method in self_methods:
            collector(method, self_object, parent=self)

        py2_self_methods = ["__getslice__", "decode"]
        for method in py2_self_methods:
            try:
                collector(method, self_object)
            except AttributeError:
                pass

        for method in ["rsplit", "split", "splitlines"]:
            collector(method, get_list(self_object), parent=self)

        super().__init__(str, collector.attributes)

    def get_doc(self):
        return str.__doc__


get_str = _create_builtin_getter(Str)
get_str_type = _create_builtin_type_getter(Str)


class BuiltinName(pynames.PyName):
    def __init__(self, pyobject):
        self.pyobject = pyobject

    def get_object(self):
        return self.pyobject

    def get_definition_location(self):
        return (None, None)


class Iterator(pyobjects.AbstractClass):
    def __init__(self, holding=None):
        super().__init__()
        self.holding = holding
        self.attributes = {
            "next": BuiltinName(BuiltinFunction(self.holding)),
            "__iter__": BuiltinName(BuiltinFunction(self)),
        }

    def get_attributes(self):
        return self.attributes

    def get_returned_object(self, args):
        return self.holding


get_iterator = _create_builtin_getter(Iterator)


class Generator(pyobjects.AbstractClass):
    def __init__(self, holding=None):
        super().__init__()
        self.holding = holding
        self.attributes = {
            "next": BuiltinName(BuiltinFunction(self.holding)),
            "__iter__": BuiltinName(BuiltinFunction(get_iterator(self.holding))),
            "close": BuiltinName(BuiltinFunction()),
            "send": BuiltinName(BuiltinFunction()),
            "throw": BuiltinName(BuiltinFunction()),
        }

    def get_attributes(self):
        return self.attributes

    def get_returned_object(self, args):
        return self.holding


get_generator = _create_builtin_getter(Generator)


class File(BuiltinClass):
    def __init__(self, filename=None, mode="r", *args):
        self.filename = filename
        self.mode = mode
        self.args = args
        str_object = get_str()
        str_list = get_list(get_str())
        attributes = {}

        def add(name, returned=None, function=None):
            builtin = getattr(io.TextIOBase, name, None)
            attributes[name] = BuiltinName(
                BuiltinFunction(returned=returned, function=function, builtin=builtin)
            )

        add("__iter__", get_iterator(str_object))
        add("__enter__", returned=pyobjects.PyObject(self))
        for method in ["next", "read", "readline", "readlines"]:
            add(method, str_list)
        for method in [
            "close",
            "flush",
            "lineno",
            "isatty",
            "seek",
            "tell",
            "truncate",
            "write",
            "writelines",
        ]:
            add(method)
        super().__init__(open, attributes)


get_file = _create_builtin_getter(File)
get_file_type = _create_builtin_type_getter(File)


class Property(BuiltinClass):
    def __init__(self, fget=None, fset=None, fdel=None, fdoc=None):
        self._fget = fget
        self._fdoc = fdoc
        attributes = {
            "fget": BuiltinName(BuiltinFunction()),
            "fset": BuiltinName(pynames.UnboundName()),
            "fdel": BuiltinName(pynames.UnboundName()),
            "__new__": BuiltinName(BuiltinFunction(function=_property_function)),
        }
        super().__init__(property, attributes)

    def get_property_object(self, args):
        if isinstance(self._fget, pyobjects.AbstractFunction):
            return self._fget.get_returned_object(args)


def _property_function(args):
    parameters = args.get_arguments(["fget", "fset", "fdel", "fdoc"])
    return pyobjects.PyObject(Property(parameters[0]))


class Lambda(pyobjects.AbstractFunction):
    def __init__(self, node, scope):
        super().__init__()
        self.node = node
        self.arguments = node.args
        self.scope = scope

    def get_returned_object(self, args):
        result = rope.base.evaluate.eval_node(self.scope, self.node.body)
        if result is not None:
            return result.get_object()
        else:
            return pyobjects.get_unknown()

    def get_module(self):
        return self.parent.get_module()

    def get_scope(self):
        return self.scope

    def get_kind(self):
        return "lambda"

    def get_ast(self):
        return self.node

    def get_attributes(self):
        return {}

    def get_name(self):
        return "lambda"

    def get_param_names(self, special_args=True):
        result = [node.arg for node in self.arguments.args if isinstance(node, ast.arg)]
        if self.arguments.vararg:
            result.append("*" + self.arguments.vararg.arg)
        if self.arguments.kwarg:
            result.append("**" + self.arguments.kwarg.arg)
        return result

    @property
    def parent(self):
        return self.scope.pyobject


class BuiltinObject(BuiltinClass):
    def __init__(self):
        super().__init__(object, {})


class BuiltinType(BuiltinClass):
    def __init__(self):
        super().__init__(type, {})


def _infer_sequence_for_pyname(pyname):
    if pyname is None:
        return None
    seq = pyname.get_object()
    args = arguments.ObjectArguments([pyname])
    if "__iter__" in seq:
        obj = seq["__iter__"].get_object()
        if not isinstance(obj, pyobjects.AbstractFunction):
            return None
        iter = obj.get_returned_object(args)
        if iter is not None and "next" in iter:
            holding = iter["next"].get_object().get_returned_object(args)
            return holding


def _create_builtin(args, creator):
    passed = args.get_pynames(["sequence"])[0]
    if passed is None:
        holding = None
    else:
        holding = _infer_sequence_for_pyname(passed)
    if holding is not None:
        return creator(holding)
    else:
        return creator()


def _open_function(args):
    return _create_builtin(args, get_file)


def _range_function(args):
    return get_list()


def _reversed_function(args):
    return _create_builtin(args, get_iterator)


def _sorted_function(args):
    return _create_builtin(args, get_list)


def _super_function(args):
    passed_class, passed_self = args.get_arguments(["type", "self"])
    if passed_self is None:
        return passed_class
    else:
        # pyclass = passed_self.get_type()
        pyclass = passed_class
        if isinstance(pyclass, pyobjects.AbstractClass):
            supers = pyclass.get_superclasses()
            if supers:
                return pyobjects.PyObject(supers[0])
        return passed_self


def _zip_function(args):
    args = args.get_pynames(["sequence"])
    objects = []
    for seq in args:
        if seq is None:
            holding = None
        else:
            holding = _infer_sequence_for_pyname(seq)
        objects.append(holding)
    tuple = get_tuple(*objects)
    return get_list(tuple)


def _enumerate_function(args):
    passed = args.get_pynames(["sequence"])[0]
    if passed is None:
        holding = None
    else:
        holding = _infer_sequence_for_pyname(passed)
    tuple = get_tuple(None, holding)
    return get_iterator(tuple)


def _iter_function(args):
    passed = args.get_pynames(["sequence"])[0]
    if passed is None:
        holding = None
    else:
        holding = _infer_sequence_for_pyname(passed)
    return get_iterator(holding)


def _input_function(args):
    return get_str()


_initial_builtins = {
    "list": BuiltinName(get_list_type()),
    "dict": BuiltinName(get_dict_type()),
    "tuple": BuiltinName(get_tuple_type()),
    "set": BuiltinName(get_set_type()),
    "str": BuiltinName(get_str_type()),
    "open": BuiltinName(BuiltinFunction(function=_open_function, builtin=open)),
    "range": BuiltinName(BuiltinFunction(function=_range_function, builtin=range)),
    "reversed": BuiltinName(
        BuiltinFunction(function=_reversed_function, builtin=reversed)
    ),
    "sorted": BuiltinName(BuiltinFunction(function=_sorted_function, builtin=sorted)),
    "super": BuiltinName(BuiltinFunction(function=_super_function, builtin=super)),
    "property": BuiltinName(
        BuiltinFunction(function=_property_function, builtin=property)
    ),
    "zip": BuiltinName(BuiltinFunction(function=_zip_function, builtin=zip)),
    "enumerate": BuiltinName(
        BuiltinFunction(function=_enumerate_function, builtin=enumerate)
    ),
    "object": BuiltinName(BuiltinObject()),
    "type": BuiltinName(BuiltinType()),
    "iter": BuiltinName(BuiltinFunction(function=_iter_function, builtin=iter)),
    "input": BuiltinName(BuiltinFunction(function=_input_function, builtin=input)),
}

builtins = BuiltinModule("builtins", initial=_initial_builtins)



"""
Functions to find importable names.

Can extract names from source code of a python file, .so object, or builtin module.
"""

import inspect
import logging
import pathlib
from importlib import import_module
from typing import Generator, List

from rope.base import ast

from .defs import (
    ModuleCompiled,
    ModuleFile,
    ModuleInfo,
    Name,
    NameType,
    Package,
    PartialName,
    Source,
)

logger = logging.getLogger(__name__)


def get_type_ast(node: ast.AST) -> NameType:
    """Get the lsp type of a node."""
    if isinstance(node, ast.ClassDef):
        return NameType.Class
    if isinstance(node, ast.FunctionDef):
        return NameType.Function
    if isinstance(node, ast.Assign):
        return NameType.Variable
    return NameType.Variable  # default value


def get_names_from_file(
    module: pathlib.Path,
    package_name: str = "",
    underlined: bool = False,
    process_imports: bool = False,
) -> Generator[PartialName, None, None]:
    """Get all the names from a given file using ast."""
    try:
        root_node = ast.parse(module.read_bytes())
    except SyntaxError as error:
        print(error)
        return
    for node in ast.iter_child_nodes(root_node):
        if isinstance(node, ast.Assign):
            for target in node.targets:
                try:
                    assert isinstance(target, ast.Name)
                    if underlined or not target.id.startswith("_"):
                        yield PartialName(
                            target.id,
                            get_type_ast(node),
                        )
                except (AttributeError, AssertionError):
                    # TODO handle tuple assignment
                    pass
        elif isinstance(node, (ast.FunctionDef, ast.ClassDef)):
            if underlined or not node.name.startswith("_"):
                yield PartialName(
                    node.name,
                    get_type_ast(node),
                )
        elif process_imports and isinstance(node, ast.ImportFrom):
            # When we process imports, we want to include names in it's own package.
            if node.level == 0:
                continue
            if not node.module or package_name is node.module.split(".")[0]:
                continue
            for name in node.names:
                if isinstance(name, ast.alias):
                    if name.asname:
                        real_name = name.asname
                    else:
                        real_name = name.name
                else:
                    real_name = name
                if underlined or not real_name.startswith("_"):
                    yield PartialName(real_name, get_type_ast(node))


def get_type_object(imported_object) -> NameType:
    """Determine the type of an object."""
    if inspect.isclass(imported_object):
        return NameType.Class
    if inspect.isfunction(imported_object) or inspect.isbuiltin(imported_object):
        return NameType.Function
    return NameType.Variable


def get_names(module: ModuleInfo, package: Package) -> List[Name]:
    """Get all names from a module and package."""
    if isinstance(module, ModuleCompiled):
        return list(
            get_names_from_compiled(package.name, package.source, module.underlined)
        )
    if isinstance(module, ModuleFile):
        return [
            combine(package, module, partial_name)
            for partial_name in get_names_from_file(
                module.filepath,
                package.name,
                underlined=module.underlined,
                process_imports=module.process_imports,
            )
        ]
    return []


def get_names_from_compiled(
    package: str,
    source: Source,
    underlined: bool = False,
) -> Generator[Name, None, None]:
    """
    Get the names from a compiled module.

    Instead of using ast, it imports the module.
    Parameters
    ----------
    package : str
        package to import. Must be in sys.path
    underlined : bool
        include underlined names
    """
    # builtins is banned because you never have to import it
    # python_crun is banned because it crashes python
    banned = ["builtins", "python_crun"]
    if package in banned or (package.startswith("_") and not underlined):
        return  # Builtins is redundant since you don't have to import it.
    if source not in (Source.BUILTIN, Source.STANDARD):
        return
    try:
        module = import_module(str(package))
    except ImportError:
        logger.error(f"{package} could not be imported for autoimport analysis")
        return
    else:
        for name, value in inspect.getmembers(module):
            if underlined or not name.startswith("_"):
                if (
                    inspect.isclass(value)
                    or inspect.isfunction(value)
                    or inspect.isbuiltin(value)
                ):
                    yield Name(
                        str(name), package, package, source, get_type_object(value)
                    )


def combine(package: Package, module: ModuleFile, name: PartialName) -> Name:
    """Combine information to form a full name."""
    return Name(name.name, module.modname, package.name, package.source, name.name_type)



"""
This module serves to convert a data structure composed of Python primitives
(dict, list, tuple, int, str, None) to JSON-serializable primitives (object,
array, number, str, null).

A core feature of this serializer is that the produced will round-trip to
identical objects when deserialized by the standard library json module.
In other words, this property always holds:

    >>> original_data = ... any JSON ...
    >>> encoded = python_to_json(original_data)
    >>> serialized = json.dumps(encoded)
    >>> decoded = json.loads(serialized)
    >>> rehydrated_data = json_to_python(decoded)

    >>> assert rehydrated_data == original_data
    >>> assert encoded == decoded

Couple challenges in straight serialization that this module helps resolve:

- json.dumps() maps both Python list and tuple to JSON array. This module
  provides two variants:

  - In version=1, this module converts Python list `[1, 2, 3]` as-is and
    converts Python tuple `(1, 2, 3)` to special object construct
    `{"$": "t", "items": [1, 2, 3]}`

  - In version=2, it is the other way around, this module converts Python tuple
    `(1, 2, 3)` as-is and converts Python list `[1, 2, 3]` to special object
    construct `{"$": "l", "items": [1, 2, 3]}`

- Python dict keys can be a tuple/dict, but JSON Object keys must be strings
  This module replaces all `dict` keys with `refid` which can be resolved using
  the `encoded["references"][refid]` lookup table. Except there's a small
  optimisation, if the dict key is a string that isn't only numeric, which is
  encoded directly into the object.

- Python dict keys cannot be another dict because it is unhashable, therefore
  there's no encoding for having objects as keys either.

- There is currently no support for floating point numbers.

Note that `json_to_python` only accepts Python objects that can be the output
of `python_to_json`, there is NO guarantee for going the other way around. This
may or may not work:

    >>> python_to_json(json_to_python(original_data)) == original_data

"""


def python_to_json(o, version=1):
    if version not in (1, 2):
        raise ValueError(f"Unexpected version {version}")
    references = []
    result = {
        "v": version,
        "data": _py2js(o, references, version=version),
        "references": references,
    }
    if not result["references"]:
        del result["references"]
    return result


def json_to_python(o):
    version = o["v"]
    if version not in (1, 2):
        raise ValueError(f"Unexpected version {version}")
    references = o.get("references", {})
    data = _js2py(o["data"], references, version)
    return data


def _py2js(o, references, version):
    if isinstance(o, (str, int)) or o is None:
        return o
    elif isinstance(o, tuple):
        if version == 1:
            return {
                "$": "t",
                "items": [_py2js(item, references, version) for item in o],
            }
        else:
            return [_py2js(item, references, version) for item in o]
    elif isinstance(o, list):
        if version == 2:
            return {
                "$": "l",
                "items": [_py2js(item, references, version) for item in o],
            }
        else:
            return [_py2js(item, references, version) for item in o]
    elif isinstance(o, dict):
        result = {}
        for pykey, pyvalue in o.items():
            if pykey == "$":
                raise ValueError('dict cannot contain reserved key "$"')
            if isinstance(pykey, str) and not pykey.isdigit():
                result[pykey] = _py2js(pyvalue, references, version)
            else:
                assert isinstance(pykey, (str, int, tuple)) or pykey is None
                assert not isinstance(pykey, list)
                refid = len(references)
                references.append(_py2js(pykey, references, version))
                result[str(refid)] = _py2js(pyvalue, references, version)
        return result
    raise TypeError(f"Object of type {type(o)} is not allowed {o}")


def _js2py(o, references, version):
    if isinstance(o, (str, int)) or o is None:
        return o
    elif isinstance(o, list):
        if version == 1:
            return list(_js2py(item, references, version) for item in o)
        elif version == 2:
            return tuple(_js2py(item, references, version) for item in o)
        raise ValueError(f"Unexpected version {version}")
    elif isinstance(o, dict):
        result = {}
        if "$" in o:
            if o["$"] == "t":
                assert version == 1
                data = o["items"]
                return tuple(_js2py(item, references, version) for item in data)
            elif o["$"] == "l":
                assert version == 2
                data = o["items"]
                return list(_js2py(item, references, version) for item in data)
            raise TypeError(f'Unrecognized object of type: {o["$"]} {o}')
        else:
            for refid, jsvalue in o.items():
                assert isinstance(refid, str)
                if refid.isdigit():
                    refid = int(refid)
                    assert 0 <= refid < len(references)
                    jskey = references[refid]
                    pyvalue = _js2py(jsvalue, references, version)
                    pykey = _js2py(jskey, references, version)
                    result[pykey] = pyvalue
                else:
                    result[refid] = _js2py(jsvalue, references, version)
        return result
    raise TypeError(f'Object of type "{type(o).__name__}" is not allowed {o}')



import pathlib

import pytest

from ropetest import testutils


@pytest.fixture
def mod1(project):
    mod1 = testutils.create_module(project, "mod1")
    yield mod1


@pytest.fixture
def mod1_path(mod1):
    yield pathlib.Path(mod1.real_path)


@pytest.fixture
def typing_path():
    import typing

    yield pathlib.Path(typing.__file__)


@pytest.fixture
def build_env_path():
    from build import env

    yield pathlib.Path(env.__file__)


@pytest.fixture
def build_path():
    import build

    # Uses __init__.py so we need the parent

    yield pathlib.Path(build.__file__).parent


@pytest.fixture
def compiled_lib():
    import _sqlite3

    yield "_sqlite3", pathlib.Path(_sqlite3.__file__)



import unittest
from textwrap import dedent

from rope.base import exceptions
from rope.contrib import generate
from ropetest import testutils


class GenerateTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, "mod1")
        self.mod2 = testutils.create_module(self.project, "mod2")
        self.pkg = testutils.create_package(self.project, "pkg")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _get_generate(self, offset):
        return generate.GenerateVariable(self.project, self.mod, offset)

    def _get_generate_class(self, offset, goal_mod=None):
        return generate.GenerateClass(
            self.project, self.mod, offset, goal_resource=goal_mod
        )

    def _get_generate_module(self, offset):
        return generate.GenerateModule(self.project, self.mod, offset)

    def _get_generate_package(self, offset):
        return generate.GeneratePackage(self.project, self.mod, offset)

    def _get_generate_function(self, offset):
        return generate.GenerateFunction(self.project, self.mod, offset)

    def test_getting_location(self):
        code = "a_var = name\n"
        self.mod.write(code)
        generator = self._get_generate(code.index("name"))
        self.assertEqual((self.mod, 1), generator.get_location())

    def test_generating_variable(self):
        code = dedent("""\
            a_var = name
        """)
        self.mod.write(code)
        changes = self._get_generate(code.index("name")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                name = None


                a_var = name
            """),
            self.mod.read(),
        )

    def test_generating_variable_inserting_before_statement(self):
        code = dedent("""\
            c = 1
            c = b
        """)
        self.mod.write(code)
        changes = self._get_generate(code.index("b")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                c = 1
                b = None


                c = b
            """),
            self.mod.read(),
        )

    def test_generating_variable_in_local_scopes(self):
        code = dedent("""\
            def f():
                c = 1
                c = b
        """)
        self.mod.write(code)
        changes = self._get_generate(code.index("b")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                def f():
                    c = 1
                    b = None
                    c = b
            """),
            self.mod.read(),
        )

    def test_generating_variable_in_other_modules(self):
        code = dedent("""\
            import mod2
            c = mod2.b
        """)
        self.mod.write(code)
        generator = self._get_generate(code.index("b"))
        self.project.do(generator.get_changes())
        self.assertEqual((self.mod2, 1), generator.get_location())
        self.assertEqual("b = None\n", self.mod2.read())

    def test_generating_variable_in_classes(self):
        code = dedent("""\
            class C(object):
                def f(self):
                    pass
            c = C()
            a_var = c.attr""")
        self.mod.write(code)
        changes = self._get_generate(code.index("attr")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                class C(object):
                    def f(self):
                        pass

                    attr = None
                c = C()
                a_var = c.attr"""),
            self.mod.read(),
        )

    def test_generating_variable_in_classes_removing_pass(self):
        code = dedent("""\
            class C(object):
                pass
            c = C()
            a_var = c.attr""")
        self.mod.write(code)
        changes = self._get_generate(code.index("attr")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                class C(object):

                    attr = None
                c = C()
                a_var = c.attr"""),
            self.mod.read(),
        )

    def test_generating_variable_in_packages(self):
        code = "import pkg\na = pkg.a\n"
        self.mod.write(code)
        generator = self._get_generate(code.rindex("a"))
        self.project.do(generator.get_changes())
        init = self.pkg.get_child("__init__.py")
        self.assertEqual((init, 1), generator.get_location())
        self.assertEqual("a = None\n", init.read())

    def test_generating_classes(self):
        code = "c = C()\n"
        self.mod.write(code)
        changes = self._get_generate_class(code.index("C")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                class C(object):
                    pass


                c = C()
            """),
            self.mod.read(),
        )

    def test_generating_classes_in_other_module(self):
        code = "c = C()\n"
        self.mod.write(code)
        changes = self._get_generate_class(code.index("C"), self.mod2).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                class C(object):
                    pass
            """),
            self.mod2.read(),
        )
        self.assertEqual(
            dedent("""\
                from mod2 import C
                c = C()
            """),
            self.mod.read(),
        )

    def test_generating_modules(self):
        code = dedent("""\
            import pkg
            pkg.mod
        """)
        self.mod.write(code)
        generator = self._get_generate_module(code.rindex("mod"))
        self.project.do(generator.get_changes())
        mod = self.pkg.get_child("mod.py")
        self.assertEqual((mod, 1), generator.get_location())
        self.assertEqual(
            dedent("""\
                import pkg.mod
                pkg.mod
            """),
            self.mod.read(),
        )

    def test_generating_packages(self):
        code = dedent("""\
            import pkg
            pkg.pkg2
        """)
        self.mod.write(code)
        generator = self._get_generate_package(code.rindex("pkg2"))
        self.project.do(generator.get_changes())
        pkg2 = self.pkg.get_child("pkg2")
        init = pkg2.get_child("__init__.py")
        self.assertEqual((init, 1), generator.get_location())
        self.assertEqual(
            dedent("""\
                import pkg.pkg2
                pkg.pkg2
            """),
            self.mod.read(),
        )

    def test_generating_function(self):
        code = "a_func()\n"
        self.mod.write(code)
        changes = self._get_generate_function(code.index("a_func")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                def a_func():
                    pass


                a_func()
            """),
            self.mod.read(),
        )

    def test_generating_modules_with_empty_primary(self):
        code = "mod\n"
        self.mod.write(code)
        generator = self._get_generate_module(code.rindex("mod"))
        self.project.do(generator.get_changes())
        mod = self.project.root.get_child("mod.py")
        self.assertEqual((mod, 1), generator.get_location())
        self.assertEqual("import mod\nmod\n", self.mod.read())

    def test_generating_variable_already_exists(self):
        code = dedent("""\
            b = 1
            c = b
        """)
        self.mod.write(code)
        with self.assertRaises(exceptions.RefactoringError):
            self._get_generate(code.index("b")).get_changes()

    def test_generating_variable_primary_cannot_be_determined(self):
        code = "c = can_not_be_found.b\n"
        self.mod.write(code)
        with self.assertRaises(exceptions.RefactoringError):
            self._get_generate(code.rindex("b")).get_changes()

    def test_generating_modules_when_already_exists(self):
        code = "mod2\n"
        self.mod.write(code)
        generator = self._get_generate_module(code.rindex("mod"))
        with self.assertRaises(exceptions.RefactoringError):
            self.project.do(generator.get_changes())

    def test_generating_static_methods(self):
        code = dedent("""\
            class C(object):
                pass
            C.a_func()
        """)
        self.mod.write(code)
        changes = self._get_generate_function(code.index("a_func")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                class C(object):

                    @staticmethod
                    def a_func():
                        pass
                C.a_func()
            """),
            self.mod.read(),
        )

    def test_generating_methods(self):
        code = dedent("""\
            class C(object):
                pass
            c = C()
            c.a_func()
        """)
        self.mod.write(code)
        changes = self._get_generate_function(code.index("a_func")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                class C(object):

                    def a_func(self):
                        pass
                c = C()
                c.a_func()
            """),
            self.mod.read(),
        )

    def test_generating_constructors(self):
        code = dedent("""\
            class C(object):
                pass
            c = C()
        """)
        self.mod.write(code)
        changes = self._get_generate_function(code.rindex("C")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                class C(object):

                    def __init__(self):
                        pass
                c = C()
            """),
            self.mod.read(),
        )

    def test_generating_calls(self):
        code = dedent("""\
            class C(object):
                pass
            c = C()
            c()
        """)
        self.mod.write(code)
        changes = self._get_generate_function(code.rindex("c")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                class C(object):

                    def __call__(self):
                        pass
                c = C()
                c()
            """),
            self.mod.read(),
        )

    def test_generating_calls_in_other_modules(self):
        self.mod2.write(dedent("""\
            class C(object):
                pass
        """))
        code = dedent("""\
            import mod2
            c = mod2.C()
            c()
        """)
        self.mod.write(code)
        changes = self._get_generate_function(code.rindex("c")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                class C(object):

                    def __call__(self):
                        pass
            """),
            self.mod2.read(),
        )

    def test_generating_function_handling_arguments(self):
        code = "a_func(1)\n"
        self.mod.write(code)
        changes = self._get_generate_function(code.index("a_func")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                def a_func(arg0):
                    pass


                a_func(1)
            """),
            self.mod.read(),
        )

    def test_generating_function_handling_keyword_xarguments(self):
        code = "a_func(p=1)\n"
        self.mod.write(code)
        changes = self._get_generate_function(code.index("a_func")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                def a_func(p):
                    pass


                a_func(p=1)
            """),
            self.mod.read(),
        )

    def test_generating_function_handling_arguments_better_naming(self):
        code = dedent("""\
            a_var = 1
            a_func(a_var)
        """)
        self.mod.write(code)
        changes = self._get_generate_function(code.index("a_func")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                a_var = 1
                def a_func(a_var):
                    pass


                a_func(a_var)
            """),
            self.mod.read(),
        )

    def test_generating_variable_in_other_modules2(self):
        self.mod2.write("\n\n\nprint(1)\n")
        code = dedent("""\
            import mod2
            c = mod2.b
        """)
        self.mod.write(code)
        generator = self._get_generate(code.index("b"))
        self.project.do(generator.get_changes())
        self.assertEqual((self.mod2, 5), generator.get_location())
        self.assertEqual(
            dedent("""\



                print(1)


                b = None
            """),
            self.mod2.read(),
        )

    def test_generating_function_in_a_suite(self):
        code = dedent("""\
            if True:
                a_func()
        """)
        self.mod.write(code)
        changes = self._get_generate_function(code.index("a_func")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                def a_func():
                    pass


                if True:
                    a_func()
            """),
            self.mod.read(),
        )

    def test_generating_function_in_a_suite_in_a_function(self):
        code = dedent("""\
            def f():
                a = 1
                if 1:
                    g()
        """)
        self.mod.write(code)
        changes = self._get_generate_function(code.index("g()")).get_changes()
        self.project.do(changes)
        self.assertEqual(
            dedent("""\
                def f():
                    a = 1
                    def g():
                        pass
                    if 1:
                        g()
            """),
            self.mod.read(),
        )

    def test_create_generate_class_with_goal_resource(self):
        code = "c = C()\n"
        self.mod.write(code)

        result = generate.create_generate(
            "class", self.project, self.mod, code.index("C"), goal_resource=self.mod2
        )

        self.assertTrue(isinstance(result, generate.GenerateClass))
        self.assertEqual(result.goal_resource, self.mod2)

    def test_create_generate_class_without_goal_resource(self):
        code = "c = C()\n"
        self.mod.write(code)

        result = generate.create_generate(
            "class", self.project, self.mod, code.index("C")
        )

        self.assertTrue(isinstance(result, generate.GenerateClass))
        self.assertIsNone(result.goal_resource)
